﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace MontoyaTech.Parallel.Net
{
    /// <summary>
    /// A set of helper functions for working with ParallelTasks.
    /// </summary>
    public class ParallelTask
    {
        /// <summary>
        /// Whether or not this parallel task is currently running.
        /// </summary>
        public bool Running
        {
            get
            {
                return this.running && !this.completed && !this.failed;
            }
        }

        /// <summary>
        /// Whether or not this parallel task completed successfully.
        /// </summary>
        public bool Completed
        {
            get
            {
                return this.completed;
            }
        }

        /// <summary>
        /// Whether or not this parallel task failed due to an exception.
        /// </summary>
        public bool Failed
        {
            get
            {
                return this.failed;
            }
        }

        /// <summary>
        /// The exception that was thrown while running this parallel task if it failed.
        /// </summary>
        public Exception ThrownException
        {
            get
            {
                return this.thrownException;
            }
        }

        protected bool running = false;

        protected bool completed = false;

        protected bool failed = false;

        protected bool exceptionStopped = false;

        protected ManualResetEvent handle;

        protected Exception thrownException = null;

        /// <summary>
        /// Takes a set of inputs and creates a set of parallel tasks and runs them returning the results. The results will be out of order.
        /// </summary>
        /// <typeparam name="InputT">The input type</typeparam>
        /// <typeparam name="ResultT">The result type</typeparam>
        /// <param name="inputs">The list of inputs to process</param>
        /// <param name="func">The function to run against each input and produce a result</param>
        /// <param name="maxTasks">The max number of tasks to run in parallel. If null the number of inputs becomes the number of tasks. Default is null.</param>
        /// <param name="timeout">If set this function will return early if the timeout is reached even if the tasks are not done. Default is null.</param>
        /// <param name="throwOnTimeout">If true an exception will be thrown if a timeout is reached if one was set. Default is true.</param>
        /// <param name="bubbleExceptions">If true, rethrows any exceptions from the running tasks. Default is true.</param>
        /// <returns>The list of results from the inputs processed.</returns>
        public static List<ResultT> ForEach<InputT, ResultT>(IList<InputT> inputs, Func<InputT, ResultT> func, int? maxTasks = null, TimeSpan? timeout = null, bool throwOnTimeout = true, bool bubbleExceptions = true)
        {
            if (inputs == null || inputs.Count == 0)
                return new List<ResultT>();

            var tasks = new List<ParallelTask<List<InputT>, List<ResultT>>>();

            if (!maxTasks.HasValue)
                maxTasks = inputs.Count;
            else if (maxTasks <= 0)
                maxTasks = 1;

            int itemsPerChunk = (inputs.Count / maxTasks.Value) + 1;

            int itemIndex = 0;

            for (int i = 0; i < maxTasks.Value; i++)
            {
                var chunk = new List<InputT>();

                for (int p = 0; p < itemsPerChunk; p++)
                {
                    chunk.Add(inputs[itemIndex++]);

                    if (itemIndex >= inputs.Count)
                        break;
                }

                if (chunk.Count > 0)
                {
                    tasks.Add(new ParallelTask<List<InputT>, List<ResultT>>(inputs =>
                    {
                        var results = new List<ResultT>();

                        if (inputs != null)
                            for (int i = 0; i < inputs.Count; i++)
                                results.Add(func(inputs[i]));

                        return results;
                    }, chunk));
                }

                if (itemIndex >= inputs.Count)
                    break;
            }

            if (tasks.Count == 0)
                return new List<ResultT>();

            var chunkResults = WhenAll(tasks, timeout, throwOnTimeout);

            if (bubbleExceptions)
                for (int i = 0; i < tasks.Count; i++)
                    if (tasks[i].Failed && tasks[i].ThrownException != null)
                        throw tasks[i].ThrownException;

            var results = new List<ResultT>();

            foreach (var chunk in chunkResults)
                if (chunk != null)
                    results.AddRange(chunk);

            return results;
        }

        /// <summary>
        /// Takes a set of inputs and creates a set of parallel tasks and runs them.
        /// </summary>
        /// <typeparam name="InputT">The input type</typeparam>
        /// <param name="inputs">The list of inputs to process</param>
        /// <param name="action">The action to run against each input</param>
        /// <param name="maxTasks">The max number of tasks to run in parallel. If null the number of inputs becomes the number of tasks. Default is null.</param>
        /// <param name="timeout">If set this function will return early if the timeout is reached even if the tasks are not done. Default is null.</param>
        /// <param name="throwOnTimeout">If true an exception will be thrown if a timeout is reached if one was set. Default is true.</param>
        /// <param name="bubbleExceptions">If true, rethrows any exceptions from the running tasks. Default is true.</param>
        public static void ForEach<InputT>(IList<InputT> inputs, Action<InputT> action, int? maxTasks = null, TimeSpan? timeout = null, bool throwOnTimeout = true, bool bubbleExceptions = true)
        {
            if (inputs == null || inputs.Count == 0)
                return;

            var tasks = new List<ParallelTask<List<InputT>>>();

            if (!maxTasks.HasValue)
                maxTasks = inputs.Count;
            else if (maxTasks <= 0)
                maxTasks = 1;

            int itemsPerChunk = (inputs.Count / maxTasks.Value) + 1;

            int itemIndex = 0;

            for (int i = 0; i < maxTasks.Value; i++)
            {
                var chunk = new List<InputT>();

                for (int p = 0; p < itemsPerChunk; p++)
                {
                    chunk.Add(inputs[itemIndex++]);

                    if (itemIndex >= inputs.Count)
                        break;
                }

                if (chunk.Count > 0)
                {
                    tasks.Add(new ParallelTask<List<InputT>>(inputs =>
                    {
                        if (inputs != null)
                            for (int i = 0; i < inputs.Count; i++)
                                action(inputs[i]);
                    }, chunk));
                }

                if (itemIndex >= inputs.Count)
                    break;
            }

            if (tasks.Count == 0)
                return;

            WhenAll(tasks, timeout, throwOnTimeout);

            if (bubbleExceptions)
                for (int i = 0; i < tasks.Count; i++)
                    if (tasks[i].Failed && tasks[i].ThrownException != null)
                        throw tasks[i].ThrownException;
        }

        /// <summary>
        /// Runs a list of tasks and waits until they have been completed or failed and returns the results.
        /// </summary>
        /// <typeparam name="InputT">The input type</typeparam>
        /// <typeparam name="ResultT">The result type</typeparam>
        /// <param name="tasks">The list of tasks to run</param>
        /// <param name="timeout">If set this function will leave early if the timeout is met. Default is null.</param>
        /// <param name="throwOnTimeout">If set an exception will be thrown if a timeout is reached. Default is true.</param>
        /// <returns>The list of results from the tasks.</returns>
        /// <exception cref="Exception">A timeout exception is thrown if throwOnTimeout is true and a timeout is reached.</exception>
        public static List<ResultT> WhenAll<InputT, ResultT>(IList<ParallelTask<InputT, ResultT>> tasks, TimeSpan? timeout = null, bool throwOnTimeout = true)
        {
            if (tasks == null || tasks.Count == 0)
                return new List<ResultT>();

            var countDown = new CountdownEvent(tasks.Count);

            foreach (var task in tasks)
                task.Run(timeout, false, throwOnTimeout, countDown);

            if (!timeout.HasValue)
                countDown.Wait();
            else if (timeout.HasValue && !countDown.Wait(timeout.Value) && throwOnTimeout)
                throw new Exception($"ParallelTasks did not complete within {timeout.Value.TotalMilliseconds} miliseconds");

            var results = new List<ResultT>();

            foreach (var task in tasks)
                results.Add(task.Result);

            return results;
        }

        /// <summary>
        /// Runs a list of tasks and waits until they have been completed or failed.
        /// </summary>
        /// <typeparam name="InputT">The input type</typeparam>
        /// <param name="tasks">The list of tasks to run</param>
        /// <param name="timeout">If set this function will leave early if the timeout is met. Default is null.</param>
        /// <param name="throwOnTimeout">If set an exception will be thrown if a timeout is reached. Default is true.</param>
        /// <exception cref="Exception">A timeout exception is thrown if throwOnTimeout is true and a timeout is reached.</exception>
        public static void WhenAll<InputT>(IList<ParallelTask<InputT>> tasks, TimeSpan? timeout = null, bool throwOnTimeout = true)
        {
            if (tasks == null || tasks.Count == 0)
                return;

            var countDown = new CountdownEvent(tasks.Count);

            foreach (var task in tasks)
                task.Run(timeout, false, throwOnTimeout, countDown);

            if (!timeout.HasValue)
                countDown.Wait();
            else if (timeout.HasValue && !countDown.Wait(timeout.Value) && throwOnTimeout)
                throw new Exception($"ParallelTasks did not complete within {timeout.Value.TotalMilliseconds} miliseconds");
        }
    }
}
