﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace MontoyaTech.Parallel.Net
{
    /// <summary>
    /// The outline of a ParallelTask that takes in an input, and produces and output when ran.
    /// </summary>
    /// <typeparam name="InputT"></typeparam>
    /// <typeparam name="ResultT"></typeparam>
    public class ParallelTask<InputT, ResultT> : ParallelTask
    {
        /// <summary>
        /// The result of this parallel task if it completed successfully.
        /// </summary>
        public ResultT Result
        {
            get
            {
                return this.result;
            }
        }

        protected ResultT result = default(ResultT);

        protected Func<InputT, ResultT> func = null;

        protected InputT input = default(InputT);

        /// <summary>
        /// Creates a new parallel task with the worker function and input.
        /// </summary>
        /// <param name="func"></param>
        /// <param name="input"></param>
        public ParallelTask(Func<InputT, ResultT> func, InputT input)
        {
            this.func = func;

            this.input = input;
        }

        /// <summary>
        /// Runs this parallel task and returns the result when it completes.
        /// </summary>
        /// <param name="timeout">Default is null, if set, this function will return if the timeout is reached regardless if the task is complete.</param>
        /// <param name="blocking">Default is true, whether or not to block the caller function and not return until this task is complete.</param>
        /// <param name="throwOnTimeout">Whether or not to throw an exception if a timeout is reached, if one was specified. Default is true.</param>
        /// <param name="countDown">An optional count down that if specified is signaled when the task completes.</param>
        /// <returns>The result from the worker function ran in this task.</returns>
        /// <exception cref="Exception">Throws a timeout exception if the task doesn't complete in time.</exception>
        public ResultT Run(TimeSpan? timeout = null, bool blocking = true, bool throwOnTimeout = true, CountdownEvent countDown = null)
        {
            this.running = true;

            this.handle = new ManualResetEvent(false);

            ThreadPool.QueueUserWorkItem((object state) =>
            {
                try
                {
                    this.result = this.func((InputT)state);

                    this.completed = true;
                }
                catch (Exception ex)
                {
                    this.failed = true;

                    this.thrownException = ex;
                }

                this.handle.Set();

                if (countDown != null)
                    countDown.Signal();

            }, this.input);

            if (blocking)
            {
                if (!timeout.HasValue)
                    this.handle.WaitOne();
                else if (!this.handle.WaitOne(timeout.Value) && throwOnTimeout)
                    throw new Exception($"ParallelTask did not complete within {timeout.Value.TotalMilliseconds} miliseconds");
            }

            return this.result;
        }
    }

    /// <summary>
    /// The outline of a ParallelTask that takes an input and processes it when ran.
    /// </summary>
    /// <typeparam name="InputT"></typeparam>
    public class ParallelTask<InputT> : ParallelTask
    {
        protected Action<InputT> action = null;

        protected InputT input = default(InputT);

        /// <summary>
        /// Creates a new parallel task with the worker function and input.
        /// </summary>
        /// <param name="action"></param>
        /// <param name="input"></param>
        public ParallelTask(Action<InputT> action, InputT input)
        {
            this.action = action;

            this.input = input;
        }

        /// <summary>
        /// Runs this parallel task and returns when it completes.
        /// </summary>
        /// <param name="timeout">Default is null, if set, this function will return if the timeout is reached regardless if the task is complete.</param>
        /// <param name="blocking">Default is true, whether or not to block the caller function and not return until this task is complete.</param>
        /// <param name="throwOnTimeout">Whether or not to throw an exception if a timeout is reached, if one was specified. Default is true.</param>
        /// <param name="countDown">An optional count down that if specified is signaled when the task completes.</param>
        /// <exception cref="Exception">Throws a timeout exception if the task doesn't complete in time.</exception>
        public void Run(TimeSpan? timeout = null, bool blocking = true, bool throwOnTimeout = true, CountdownEvent countDown = null)
        {
            this.running = true;

            this.handle = new ManualResetEvent(false);

            ThreadPool.QueueUserWorkItem((object state) =>
            {
                try
                {
                    this.action((InputT)state);

                    this.completed = true;
                }
                catch (Exception ex)
                {
                    this.failed = true;

                    this.thrownException = ex;
                }

                this.handle.Set();

                if (countDown != null)
                    countDown.Signal();

            }, this.input);

            if (blocking)
            {
                if (!timeout.HasValue)
                    this.handle.WaitOne();
                else if (!this.handle.WaitOne(timeout.Value) && throwOnTimeout)
                    throw new Exception($"ParallelTask did not complete within {timeout.Value.TotalMilliseconds} miliseconds");
            }
        }
    }
}
